# author: Kwang-Hyun Park (akaii@kw.ac.kr)

from named_element import NamedElement

DeviceType = type('Enum', (), { 'SENSOR': 0, 'EFFECTOR': 1, 'COMMAND': 2, 'EVENT': 3 })
DataType = type('Enum', (), { 'INTEGER': 4, 'FLOAT': 5, 'STRING': 6 })

class Device(NamedElement):
	def __init__(self, uid, name, deviceType, dataType, dataSize, initialValue, minValue, maxValue):
		NamedElement.__init__(self, name)
		self.__uid = uid
		self.__deviceType = deviceType
		self.__dataType = dataType
		self.__dataSize = dataSize
		self.__initialValue = initialValue
		self.__minValue = minValue
		self.__maxValue = maxValue
		if dataSize > 0:
			self.__data = [initialValue] * dataSize
		else:
			self.__data = []
		self.__fired = False
		self.__event = False
		self.__written = False
		self.__canNotify = False
		self.__deviceDataChangedListeners = []

	def __getUid(self):
		return self.__uid

	def getId(self):
		return self.__uid & 0xfff00fff

	def getProductId(self):
		return (self.__uid & 0x000ff000) >> 12

	def getDeviceType(self):
		return self.__deviceType

	def getDataType(self):
		return self.__dataType

	def getDataSize(self):
		return self.__dataSize

	def __isWritten(self):
		return self.__written

	def __clearWritten(self):
		self.__written = False

	def e(self):
		return self.__event

	def __checkDataType(self, value):
		if self.__dataType == DataType.INTEGER or self.__dataType == DataType.FLOAT:
			return isinstance(value, (int, float))
		elif self.__dataType == DataType.STRING:
			return isinstance(value, str)
		return False

	def __checkRange(self, value):
		if isinstance(value, (int, float)):
			if value < self.__minValue: value = self.__minValue
			elif value > self.__maxValue: value = self.__maxValue
			if self.__dataType == DataType.INTEGER:
				return int(value)
			else:
				return float(value)
		return str(value)

	def read(self, arg=None):
		thisData = self.__data
		thisLen = len(thisData)
		if isinstance(arg, (int, float)):
			index = int(arg)
			if index >= 0 and index < thisLen:
				return thisData[index]
		elif isinstance(arg, list):
			length = min(thisLen, len(arg))
			arg[:length] = thisData[:length]
			return length
		elif thisLen > 0:
			return thisData[0]
		if self.__dataType == DataType.STRING:
			return ''
		else:
			return 0

	def write(self, arg1, arg2=None):
		if self.__deviceType == DeviceType.SENSOR or self.__deviceType == DeviceType.EVENT:
			return False
		thisData = self.__data
		if isinstance(arg1, (int, float)) and isinstance(arg2, (int, float, str)):
			index = int(arg1)
			if index < 0:
				return False
			if self.__checkDataType(arg2) == False:
				return False
			thisLen = len(thisData)
			if index >= thisLen:
				if self.__dataSize < 0:
					newData = [self.__initialValue] * (index + 1)
					newData[:thisLen] = thisData
					thisData = self.__data = newData
				else:
					return False
			thisData[index] = self.__checkRange(arg2)
			self.__fired = True
			self.__written = True
			self.__canNotify = True
			return True
		elif isinstance(arg1, (int, float, str)):
			if self.__checkDataType(arg1) == False:
				return False
			if len(thisData) <= 0:
				if self.__dataSize < 0:
					thisData = self.__data = [0]
				else:
					return False
			thisData[0] = self.__checkRange(arg1)
			self.__fired = True
			self.__written = True
			self.__canNotify = True
			return True
		elif isinstance(arg1, (list, tuple)):
			dataLen = len(arg1)
			if dataLen <= 0:
				return 0
			if self.__dataSize < 0:
				if len(thisData) != dataLen:
					thisData = self.__data = [self.__initialValue] * dataLen
			thisLen = len(thisData)
			if thisLen <= 0:
				return 0
			length = min(thisLen, dataLen)
			for i in range(length):
				if self.__checkDataType(arg1[i]):
					thisData[i] = self.__checkRange(arg1[i])
					self.__fired = True
					self.__written = True
					self.__canNotify = True
			return length
		return False

	def __put(self, value):
		self.__data[0] = value
		self.__fired = True
		self.__canNotify = True

	def __putAt(self, index, value):
		self.__data[index] = value
		self.__fired = True
		self.__canNotify = True

	def __reset(self):
		self.__fired = False
		self.__event = False
		self.__written = False
		if self.__dataSize > 0:
			self.__data = [self.__initialValue] * self.__dataSize
		else:
			self.__data = []

	def addDeviceDataChangedListener(self, listener):
		if listener:
			self.__deviceDataChangedListeners.append(listener)

	def removeDeviceDataChangedListener(self, listener):
		if listener:
			self.__deviceDataChangedListeners.remove(listener)

	def clearDeviceDataChangedListeners(self):
		self.__deviceDataChangedListeners = []

	def __updateDeviceState(self):
		self.__event = self.__fired
		self.__fired = False

	def __notifyDeviceDataChanged(self):
		if len(self.__deviceDataChangedListeners) > 0:
			canNotify = True
			if self.__deviceType == DeviceType.COMMAND or self.__deviceType == DeviceType.EVENT:
				canNotify = self.__canNotify
			if canNotify:
				data = tuple(self.__data)
				for listener in self.__deviceDataChangedListeners:
					listener(self, data)
		self.__canNotify = False