# author: Kwang-Hyun Park (akaii@kw.ac.kr)

import serial.tools.list_ports
import io

_VALID_PACKET_LENGTH = 54
_RETRY = 10
_STATE_CONNECTING = 1
_STATE_CONNECTED = 2
_STATE_CONNECTION_LOST = 3
_STATE_DISCONNECTED = 4
_STATE_DISPOSED = 5

class HamsterConnector:
	def __init__(self, id):
		self.__id = id
		self.__serial = None
		self.__address = '000000000000'
		self.__portName = ''
		self.__found = False
		self.__connected = False

	def open(self):
		ports = filter(lambda x: x[1].find('Silicon Labs') != -1, list(serial.tools.list_ports.comports()))
		for port in ports:
			try:
				s = serial.Serial(port[0], 115200, rtscts=True)
				self.__io = io.TextIOWrapper(io.BufferedRWPair(s, s, 1), newline = '\r', line_buffering = True)
				self.__portName = port[0]
				if self.__checkPort(s):
					self.__serial = s
					return
				s.close()
			except (OSError, serial.SerialException):
				pass
		self.__showMessage('No available USB to BLE bridge')

	def close(self):
		self.__connected = False
		if self.__serial != None:
			self.__serial.close()
			self.__serial = None

	def isConnected(self):
		return self.__connected

	def getAddress(self):
		return self.__address

	def __setAddress(self, address):
		self.__address = address

	def __setConnectionState(self, state):
		self.__connected = (state == _STATE_CONNECTED)
		if self.__found == False and self.__connected:
			self.__found = True
		if self.__found:
			if state == _STATE_CONNECTED:
				self.__showMessage('Connected: ' + self.__portName + ' ' + self.__address)
			elif state == _STATE_CONNECTION_LOST:
				self.__showMessage('Connection lost')

	def __readPacket(self, serial, startByte=None):
		try:
			packet = self.__io.readline()
			if startByte == None:
				return packet
			if packet[:2] == startByte:
				return packet
			return None
		except:
			return None

	def write(self, packet):
		if self.__serial:
			self.__io.write(packet)

	def read(self):
		if self.__serial:
			try:
				packet = self.__io.readline()
				if len(packet) == _VALID_PACKET_LENGTH:
					if self.__found == False:
						self.__checkConnection(self.__serial)
					elif self.__connected == False:
						self.__setConnectionState(_STATE_CONNECTED)
					return packet
			except:
				if self.__connected:
					self.__setConnectionState(_STATE_CONNECTION_LOST)
		return None

	def __checkPort(self, serial):
		packet1 = self.__readPacket(serial)
		packet2 = self.__readPacket(serial)
		if packet2:
			if len(packet2) == _VALID_PACKET_LENGTH:
				return self.__checkConnection(serial)
			elif packet1 and len(packet2) == 2:
				self.__showMessage('Not connected')
				return True
		return False

	def __checkConnection(self, serial):
		for i in range(_RETRY):
			self.__io.write(u'FF\r')
			packet = self.__readPacket(serial, 'FF')
			if packet:
				info = packet.split(',')
				if info and len(info) >= 5:
					if info[1] == 'Hamster' and info[2] == '04':
						self.__setAddress(info[4])
						self.__setConnectionState(_STATE_CONNECTED)
						return True
				return False
		return False

	def __showMessage(self, message):
		print 'Hamster[%d]' % self.__id, message